"""Basic Zernike functions."""

from prysm import mathops as m
from prysm.conf import config


# See JCW - http://wp.optics.arizona.edu/jcwyant/wp-content/uploads/sites/13/2016/08/ZernikePolynomialsForTheWeb.pdf


def piston(rho, phi):
    return m.ones(rho.shape)


def tip(rho, phi):
    return rho * m.cos(phi)


def tilt(rho, phi):
    return rho * m.sin(phi)


def defocus(rho, phi):
    return 2 * rho**2 - 1


def primary_astigmatism_00(rho, phi):
    return rho**2 * m.cos(2 * phi)


def primary_astigmatism_45(rho, phi):
    return rho**2 * m.sin(2 * phi)


def primary_coma_y(rho, phi):
    return (-2 * rho + 3 * rho**3) * m.cos(phi)


def primary_coma_x(rho, phi):
    return (-2 * rho + 3 * rho**3) * m.sin(phi)


def primary_spherical(rho, phi):
    return 6 * rho**4 - 6 * rho**2 + 1


def primary_trefoil_y(rho, phi):
    return rho**3 * m.cos(3 * phi)


def primary_trefoil_x(rho, phi):
    return rho**3 * m.sin(3 * phi)


def secondary_astigmatism_00(rho, phi):
    return (-3 * rho**2 + 4 * rho**4) * m.cos(2 * phi)


def secondary_astigmatism_45(rho, phi):
    return (-3 * rho**2 + 4 * rho**4) * m.sin(2 * phi)


def secondary_coma_y(rho, phi):
    return (3 * rho - 12 * rho**3 + 10 * rho**5) * m.cos(phi)


def secondary_coma_x(rho, phi):
    return (3 * rho - 12 * rho**3 + 10 * rho**5) * m.sin(phi)


def secondary_spherical(rho, phi):
    return 20 * rho**6 + - 30 * rho**4 + 12 * rho**2 - 1


def primary_tetrafoil_y(rho, phi):
    return rho**4 * m.cos(4 * phi)


def primary_tetrafoil_x(rho, phi):
    return rho**4 * m.sin(4 * phi)


def secondary_trefoil_y(rho, phi):
    return (5 * rho**5 - 4 * rho**3) * m.cos(3 * phi)


def secondary_trefoil_x(rho, phi):
    return (5 * rho**5 - 4 * rho**3) * m.sin(3 * phi)


def tertiary_astigmatism_00(rho, phi):
    return (6 * rho**2 - 20 * rho**4 + 15 * rho**6) * m.cos(2 * phi)


def tertiary_astigmatism_45(rho, phi):
    return (6 * rho**2 - 20 * rho**4 + 15 * rho**6) * m.sin(2 * phi)


def tertiary_coma_y(rho, phi):
    return (-4 * rho + 30 * rho**3 - 60 * rho**5 + 35 * rho**7) * m.cos(phi)


def tertiary_coma_x(rho, phi):
    return (-4 * rho + 30 * rho**3 - 60 * rho**5 + 35 * rho**7) * m.sin(phi)


def tertiary_spherical(rho, phi):
    return 70 * rho**8 - 140 * rho**6 + 90 * rho**4 - 20 * rho**2 + 1


def primary_pentafoil_y(rho, phi):
    return rho**5 * m.cos(5 * phi)


def primary_pentafoil_x(rho, phi):
    return rho**5 * m.sin(5 * phi)


def secondary_tetrafoil_y(rho, phi):
    return (6 * rho**6 - 5 * rho**4) * m.cos(4 * phi)


def secondary_tetrafoil_x(rho, phi):
    return (6 * rho**6 - 5 * rho**4) * m.sin(4 * phi)


def tertiary_trefoil_y(rho, phi):
    return (10 * rho**3 - 30 * rho**5 + 21 * rho**7) * m.cos(3 * phi)


def tertiary_trefoil_x(rho, phi):
    return (10 * rho**3 - 30 * rho**5 + 21 * rho**7) * m.cos(3 * phi)


def quaternary_astigmatism_00(rho, phi):
    return (10 * rho**2 - 30 * rho**4 + 21 * rho**6) * m.cos(2 * phi)


def quaternary_astigmatism_45(rho, phi):
    return (10 * rho**2 - 30 * rho**4 + 21 * rho**6) * m.sin(2 * phi)


def quaternary_coma_y(rho, phi):
    return (5 * rho - 60 * rho**3 + 210 * rho**5 - 280 * rho**7 + 126 * rho**9)\
        * m.cos(phi)


def quaternary_coma_x(rho, phi):
    return (5 * rho - 60 * rho**3 + 210 * rho**5 - 280 * rho**7 + 126 * rho**9)\
        * m.sin(phi)


def quaternary_spherical(rho, phi):
    return 252 * rho**10 \
        - 630 * rho**8 \
        + 560 * rho**6 \
        - 210 * rho**4 \
        + 30 * rho**2 \
        - 1


def primary_hexafoil_y(rho, phi):
    return rho**6 * m.cos(6 * phi)


def primary_hexafoil_x(rho, phi):
    return rho**6 * m.sin(6 * phi)


def secondary_pentafoil_y(rho, phi):
    return (7 * rho**7 - 6 * rho**5) * m.cos(5 * phi)


def secondary_pentafoil_x(rho, phi):
    return (7 * rho**7 - 6 * rho**5) * m.sin(5 * phi)


def tertiary_tetrafoil_y(rho, phi):
    return (28 * rho**8 - 42 * rho**6 + 15 * rho**4) * m.cos(4 * phi)


def tertiary_tetrafoil_x(rho, phi):
    return (28 * rho**8 - 42 * rho**6 + 15 * rho**4) * m.sin(4 * phi)


def quaternary_trefoil_y(rho, phi):
    return (84 * rho**9 - 168 * rho**7 + 105 * rho**5 - 20 * rho**3) * m.cos(3 * phi)


def quaternary_trefoil_x(rho, phi):
    return (84 * rho**9 - 168 * rho**7 + 105 * rho**5 - 20 * rho**3) * m.sin(3 * phi)


def quinternary_astigmatism_00(rho, phi):
    return (210 * rho**10 - 504 * rho**8 + 420 * rho**6 - 140 * rho**4 + 15 * rho**2) \
        * m.cos(2 * phi)


def quinternary_astigmatism_45(rho, phi):
    return (210 * rho**10 - 504 * rho**8 + 420 * rho**6 - 140 * rho**4 + 15 * rho**2) \
        * m.sin(2 * phi)


def quinternary_coma_y(rho, phi):
    return (462 * rho**11 - 1260 * rho**9 + 1260 * rho**7 - 560 * rho**5 + 105 * rho**3 - 6 * rho) \
        * m.cos(phi)


def quinternary_coma_x(rho, phi):
    return (462 * rho**11 - 1260 * rho**9 + 1260 * rho**7 - 560 * rho**5 + 105 * rho**3 - 6 * rho) \
        * m.sin(phi)


def quinternary_spherical(rho, phi):
    return 924 * rho**12 \
        - 2772 * rho**10 \
        + 3150 * rho**8 \
        - 1680 * rho**6 \
        + 420 * rho**4 \
        - 42 * rho**2 \
        + 1


# norms
piston.norm = 1
tip.norm = 2
tilt.norm = 2
defocus.norm = m.sqrt(3)
primary_astigmatism_00.norm = m.sqrt(6)
primary_astigmatism_45.norm = m.sqrt(6)
primary_coma_y.norm = 2 * m.sqrt(2)
primary_coma_x.norm = 2 * m.sqrt(2)
primary_spherical.norm = m.sqrt(5)
primary_trefoil_y.norm = 2 * m.sqrt(2)
primary_trefoil_x.norm = 2 * m.sqrt(2)
secondary_astigmatism_00.norm = m.sqrt(10)
secondary_astigmatism_45.norm = m.sqrt(10)
secondary_coma_y.norm = 2 * m.sqrt(3)
secondary_coma_x.norm = 2 * m.sqrt(3)
secondary_spherical.norm = m.sqrt(7)
primary_tetrafoil_y.norm = m.sqrt(10)
primary_tetrafoil_x.norm = m.sqrt(10)
secondary_trefoil_y.norm = 2 * m.sqrt(3)
secondary_trefoil_x.norm = 2 * m.sqrt(3)
tertiary_astigmatism_00.norm = m.sqrt(14)
tertiary_astigmatism_45.norm = m.sqrt(14)
tertiary_coma_y.norm = 4
tertiary_coma_x.norm = 4
tertiary_spherical.norm = 3
primary_pentafoil_y.norm = 2 * m.sqrt(3)
primary_pentafoil_x.norm = 2 * m.sqrt(3)
secondary_tetrafoil_y.norm = m.sqrt(14)
secondary_tetrafoil_x.norm = m.sqrt(14)
tertiary_trefoil_y.norm = 4
tertiary_trefoil_x.norm = 4
quaternary_astigmatism_00.norm = 3 * m.sqrt(2)
quaternary_astigmatism_45.norm = 3 * m.sqrt(2)
quaternary_coma_y.norm = 2 * m.sqrt(5)
quaternary_coma_x.norm = 2 * m.sqrt(5)
quaternary_spherical.norm = m.sqrt(11)
primary_hexafoil_y.norm = m.sqrt(14)
primary_hexafoil_x.norm = m.sqrt(14)
secondary_pentafoil_y.norm = 4
secondary_pentafoil_x.norm = 4
tertiary_tetrafoil_y.norm = 3 * m.sqrt(2)
tertiary_tetrafoil_x.norm = 3 * m.sqrt(2)
quaternary_trefoil_y.norm = 2 * m.sqrt(5)
quaternary_trefoil_x.norm = 2 * m.sqrt(5)
quinternary_astigmatism_00.norm = m.sqrt(22)
quinternary_astigmatism_45.norm = m.sqrt(22)
quinternary_coma_y.norm = 2 * m.sqrt(6)
quinternary_coma_x.norm = 2 * m.sqrt(6)
quinternary_spherical.norm = m.sqrt(13)

# names
piston.name = 'Piston'
tip.name = 'Tilt Y'
tilt.name = 'Tilt X'
defocus.name = 'Defocus'
primary_astigmatism_00.name = 'Primary Astigmatism 0°'
primary_astigmatism_45.name = 'Primary Astigmatism 45°'
primary_coma_y.name = 'Primary Coma Y'
primary_coma_x.name = 'Primary Coma X'
primary_spherical.name = 'Primary Spherical'
primary_trefoil_y.name = 'Primary Trefoil Y'
primary_trefoil_x.name = 'Primary Trefoil X'
secondary_astigmatism_00.name = 'Secondary Astigmatism 0°'
secondary_astigmatism_45.name = 'Secondary Astigmatism 45°'
secondary_coma_y.name = 'Secondary Coma Y'
secondary_coma_x.name = 'Secondary Coma X'
secondary_spherical.name = 'Secondary Spherical'
primary_tetrafoil_y.name = 'Primary Tetrafoil Y'
primary_tetrafoil_x.name = 'Primary Tetrafoil X'
secondary_trefoil_y.name = 'Secondary Trefoil Y'
secondary_trefoil_x.name = 'Secondary Trefoil X'
tertiary_astigmatism_00.name = 'Tertiary Astigmatism 0°'
tertiary_astigmatism_45.name = 'Tertiary Astigmatism 45°'
tertiary_coma_y.name = 'Tertiary Coma Y'
tertiary_coma_x.name = 'Tertiary Coma X'
tertiary_spherical.name = 'Tertiary Spherical'
primary_pentafoil_y.name = 'Primary Pentafoil Y'
primary_pentafoil_x.name = 'Primary Pentafoil X'
secondary_tetrafoil_y.name = 'Secondary Tetrafoil Y'
secondary_tetrafoil_x.name = 'Secondary Tetrafoil X'
tertiary_trefoil_y.name = 'Tertiary Trefoil Y'
tertiary_trefoil_x.name = 'Tertiary Trefoil X'
quaternary_astigmatism_00.name = 'Quaternary Astigmatism 0°'
quaternary_astigmatism_45.name = 'Quaternary Astigmatism 45°'
quaternary_coma_y.name = 'Quaternary Coma Y'
quaternary_coma_x.name = 'Quaternary Coma X'
quaternary_spherical.name = 'Quaternary Spherical'
primary_hexafoil_y.name = 'Primary Hexafoil Y'
primary_hexafoil_x.name = 'Primary Hexafoil X'
secondary_pentafoil_y.name = 'Secondary Pentafoil Y'
secondary_pentafoil_x.name = 'Secondary Pentafoil X'
tertiary_tetrafoil_y.name = 'Tertiary Tetrafoil Y'
tertiary_tetrafoil_x.name = 'Tertiary Tetrafoil X'
quaternary_trefoil_y.name = 'Quaternary Trefoil Y'
quaternary_trefoil_x.name = 'Quaternary Trefoil X'
quinternary_astigmatism_00.name = 'Quinternary Astigmatism 0°'
quinternary_astigmatism_45.name = 'Quinternary Astigmatism 45°'
quinternary_coma_y.name = 'Quinternary Coma Y'
quinternary_coma_x.name = 'Quinternary Coma X'
quinternary_spherical.name = 'Quinternary Spherical'

_zernikes = [
    piston,
    tip,
    tilt,
    defocus,
    primary_astigmatism_00,
    primary_astigmatism_45,
    primary_coma_y,
    primary_coma_x,
    primary_spherical,
    primary_trefoil_y,
    primary_trefoil_x,
    secondary_astigmatism_00,
    secondary_astigmatism_45,
    secondary_coma_y,
    secondary_coma_x,
    secondary_spherical,
    primary_tetrafoil_y,
    primary_tetrafoil_x,
    secondary_trefoil_y,
    secondary_trefoil_x,
    tertiary_astigmatism_00,
    tertiary_astigmatism_45,
    tertiary_coma_y,
    tertiary_coma_x,
    tertiary_spherical,
    primary_pentafoil_y,
    primary_pentafoil_x,
    secondary_tetrafoil_y,
    secondary_tetrafoil_x,
    tertiary_trefoil_y,
    tertiary_trefoil_x,
    quaternary_astigmatism_00,
    quaternary_astigmatism_45,
    quaternary_coma_y,
    quaternary_coma_x,
    quaternary_spherical,
    primary_hexafoil_y,
    primary_hexafoil_x,
    secondary_pentafoil_y,
    secondary_pentafoil_x,
    tertiary_tetrafoil_y,
    tertiary_tetrafoil_x,
    quaternary_trefoil_y,
    quaternary_trefoil_x,
    quinternary_astigmatism_00,
    quinternary_astigmatism_45,
    quinternary_coma_y,
    quinternary_coma_x,
    quinternary_spherical
]


# need to disable this for now, numba doesn't preserve function attrs with vectorize
# zernikes_cpu = [m.vectorize(func) for func in _zernikes[1:]]  # numba compiled zernikes
# zernikes_cpu.insert(0, m.jit(_zernikes[0]))
zernikes_cpu = [m.jit(_zernikes[0])]
for func in _zernikes[1:]:
    compfunc = m.vectorize(func)
    compfunc.name = func.name
    compfunc.norm = func.norm
    zernikes_cpu.append(compfunc)

zernikes_gpu = [m.fuse(func) for func in _zernikes[1:]]  # cupy compiled zernikes
zernikes_gpu.insert(0, _zernikes[0])


def change_backend(to):
    if to == 'cu':
        globals()['zernikes'] = zernikes_gpu
    elif to == 'np':
        globals()['zernikes'] = zernikes_cpu


config.chbackend_observers.append(change_backend)
config.backend = config.backend  # trigger import of math functions



"""A base optical transfer function interface."""
from scipy import interpolate

from .psf import PSF
from .fttools import forward_ft_unit
from .util import share_fig_ax
from .coordinates import polar_to_cart, uniform_cart_to_polar

from prysm import mathops as m


class MTF(object):
    """Modulation Transfer Function.

    Properties
    ----------
    tan: slice along the X axis of the MTF object.

    sag: slice along the Y axis of the MTF object.

    Attributes
    ----------
    center_x : `int`
        x center pixel location (MTF == 1 here by definition)
    center_y : `int`
        y center pixel location (MTF == 1 here by definition)
    data : `numpy.ndarray`
        MTF data array, 2D
    interpf_2d : `scipy.interpolate.RegularGridInterpolator`
        2D interpolation function used to compute exact values of the 2D MTF
    interpf_sag : `scipy.interpolate.interp1d`
        1D interpolation function used to compute exact values of the sagittal MTF
    interpf_tan : `scipy.interpolate.interp1d`
        1D interpolation function used to compute exact values of the tangential MTF
    unit_x : `numpy.ndarray`
        ordinate x coordinates
    unit_y : TYPE
        ordinate y coordinates

    """
    def __init__(self, data, unit_x, unit_y=None):
        """Create an MTF object.

        Parameters
        ----------
        data : `numpy.ndarray`
            MTF values on 2D grid
        unit_x : `numpy.ndarray`
            array of x units, 1D
        unit_y : `numpy.ndarray`
            array of y units, 1D

        """
        if unit_y is None:
            unit_y = unit_x
        self.data = data
        self.unit_x = unit_x
        self.unit_y = unit_y
        self.samples_y, self.samples_x = data.shape
        self.center_x = self.samples_x // 2
        self.center_y = self.samples_y // 2

        self.interpf_2d = None
        self.interpf_tan = None
        self.interpf_sag = None

    # quick-access slices ------------------------------------------------------

    @property
    def tan(self):
        """Retrieve the tangential MTF.

        Notes
        -----
        Assumes the object is extended in y.  If the object is extended
        along a different azimuth, this will not return the tangential MTF.

        Returns
        -------
        self.unit_x : `numpy.ndarray`
            ordinate
        self.data : `numpy.ndarray`
            coordiante

        """
        return self.unit_x[self.center_x:], self.data[self.center_y:, self.center_x]

    @property
    def sag(self):
        """Retrieve the sagittal MTF.

        Notes
        -----
        Assumes the object is extended in y.  If the object is extended along a different
        azimuth, this will not return the sagittal MTF.

        Returns
        -------
        self.unit_x : `numpy.ndarray`
            ordinate
        self.data : `numpy.ndarray`
            coordiante

        """
        return self.unit_y[self.center_y:], self.data[self.center_y, self.center_x:]

    def exact_polar(self, freqs, azimuths=None):
        """Retrieve the MTF at the specified frequency-azimuth pairs.

        Parameters
        ----------
        freqs : iterable
            radial frequencies to retrieve MTF for
        azimuths : iterable
            corresponding azimuths to retrieve MTF for

        Returns
        -------
        `list`
            MTF at the given points

        """
        self._make_interp_function_2d()

        # handle user-unspecified azimuth
        if azimuths is None:
            if type(freqs) in (int, float):
                # single azimuth
                azimuths = 0
            else:
                azimuths = [0] * len(freqs)
        # handle single azimuth, multiple freqs
        elif type(azimuths) in (int, float):
            azimuths = [azimuths] * len(freqs)

        azimuths = m.radians(azimuths)
        # handle single value case
        if type(freqs) in (int, float):
            x, y = polar_to_cart(freqs, azimuths)
            return float(self.interpf_2d((x, y), method='linear'))

        outs = []
        for freq, az in zip(freqs, azimuths):
            x, y = polar_to_cart(freq, az)
            outs.append(float(self.interpf_2d((x, y), method='linear')))
        return m.asarray(outs)

    def exact_xy(self, x, y=None):
        """Retrieve the MTF at the specified X-Y frequency pairs.

        Parameters
        ----------
        x : iterable
            X frequencies to retrieve the MTF at
        y : iterable
            Y frequencies to retrieve the MTF at

        Returns
        -------
        `list`
            MTF at the given points

        """
        self._make_interp_function_2d()

        # handle data along just x
        if y is None:
            if type(x) in (int, float):
                # single azimuth
                y = 0
            else:
                y = [0] * len(x)

        elif type(y) in (int, float):
            y = [y] * len(x)

        # handle data just along y
        if type(x) in (int, float):
            x = [x] * len(y)

        x, y = m.asarray(x), m.asarray(y)
        outs = []
        for x, y in zip(x, y):
            outs.append(float(self.interpf_2d((x, y), method='linear')))
        return m.asarray(outs)

    def exact_tan(self, freq):
        """Return data at an exact x coordinate along the y=0 axis.

        Parameters
        ----------
        freq : `number` or `numpy.ndarray`
            frequency or frequencies to return

        Returns
        -------
        `numpy.ndarray`
            ndarray of MTF values

        """
        self._make_interp_function_tansag()
        return self.interpf_tan(freq)

    def exact_sag(self, freq):
        """Return data at an exact y coordinate along the x=0 axis.

        Parameters
        ----------
        freq : `number` or `numpy.ndarray`
            frequency or frequencies to return

        Returns
        -------
        `numpy.ndarray`
            ndarray of MTF values

        """
        self._make_interp_function_tansag()
        return self.interpf_sag(freq)

    def azimuthal_average(self):
        """Return the azimuthally averaged MTF.

        Returns
        -------
        nu : `numpy.ndarray`
            spatial frequencies
        mtf : `numpy.ndarray`
            mtf values

        """
        nu, theta, mtf = uniform_cart_to_polar(self.unit_x, self.unit_y, self.data)
        l = len(nu) // 2
        return nu[l:], mtf.mean(axis=0)

    # quick-access slices ------------------------------------------------------

    # plotting -----------------------------------------------------------------

    def plot2d(self, max_freq=200, power=1, fig=None, ax=None):
        """Create a 2D plot of the MTF.

        Parameters
        ----------
        max_freq : `float`
            Maximum frequency to plot to.  Axis limits will be ((-max_freq, max_freq), (-max_freq, max_freq)).
        power : `float`
            inverse of power to stretch the MTF to/by, e.g. power=2 will plot MTF^(1/2)
        fig : `matplotlib.figure.Figure`, optional:
            Figure to draw plot in
        ax : `matplotlib.axes.Axis`, optional:
            Axis to draw plot in

        Returns
        -------
        fig : `matplotlib.figure.Figure`
            Figure to draw plot in
        ax : `matplotlib.axes.Axis`
            Axis to draw plot in

        """
        from matplotlib import colors

        left, right = self.unit_x[0], self.unit_x[-1]
        bottom, top = self.unit_y[0], self.unit_y[-1]

        fig, ax = share_fig_ax(fig, ax)

        im = ax.imshow(self.data,
                       extent=[left, right, bottom, top],
                       origin='lower',
                       cmap='Greys_r',
                       clim=(0, 1),
                       norm=colors.PowerNorm(1/power),
                       interpolation='lanczos')
        cb = fig.colorbar(im, label='MTF [Rel 1.0]', ax=ax, fraction=0.046)
        cb.outline.set_edgecolor('k')
        cb.outline.set_linewidth(0.5)
        ax.set(xlabel=r'$\nu_x$ [cy/mm]',
               ylabel=r'$\nu_y$ [cy/mm]',
               xlim=(-max_freq, max_freq),
               ylim=(-max_freq, max_freq))
        return fig, ax

    def plot_tan_sag(self, max_freq=200, fig=None, ax=None, labels=('Tangential', 'Sagittal')):
        """Create a plot of the tangential and sagittal MTF.

        Parameters
        ----------
        max_freq : `float`
            Maximum frequency to plot to.  Axis limits will be ((-max_freq, max_freq), (-max_freq, max_freq))
        fig : `matplotlib.figure.Figure`, optional:
            Figure to draw plot in
        ax : `matplotlib.axes.Axis`, optional:
            Axis to draw plot in
        labels : `iterable`
            set of labels for the two lines that will be plotted

        Returns
        -------
        fig : `matplotlib.figure.Figure`
            Figure to draw plot in
        ax : `matplotlib.axes.Axis`
            Axis to draw plot in

        """
        ut, tan = self.tan
        us, sag = self.sag

        fig, ax = share_fig_ax(fig, ax)
        ax.plot(ut, tan, label=labels[0], linestyle='-', lw=3)
        ax.plot(us, sag, label=labels[1], linestyle='--', lw=3)
        ax.set(xlabel='Spatial Frequency [cy/mm]',
               ylabel='MTF [Rel 1.0]',
               xlim=(0, max_freq),
               ylim=(0, 1))
        ax.legend(loc='lower left')
        return fig, ax

    # plotting -----------------------------------------------------------------

    # helpers ------------------------------------------------------------------

    def _make_interp_function_2d(self):
        """Generate a 2D interpolation function for this instance of MTF, used to procure MTF at exact frequencies.

        Returns
        -------
        `scipy.interpolate.RegularGridInterpolator`, this instance of an MTF object.

        """
        if self.interpf_2d is None:
            self.interpf_2d = interpolate.RegularGridInterpolator((self.unit_x, self.unit_y), self.data)

        return self.interpf_2d

    def _make_interp_function_tansag(self):
        """Generate two interpolation functions for tangential and sagittal MTF.

        Returns
        -------
        self.interpf_tan : `scipy.interpolate.interp1d`
            tangential interpolator
        self.interpf_sag : `scipy.interpolate.interp1d`
            sagittal interpolator

        """
        if self.interpf_tan is None or self.interpf_sag is None:
            ut, tan = self.tan
            us, sag = self.sag

            self.interpf_tan = interpolate.interp1d(ut, tan)
            self.interpf_sag = interpolate.interp1d(us, sag)

        return self.interpf_tan, self.interpf_sag

    @staticmethod
    def from_psf(psf):
        """Generate an MTF from a PSF.

        Parameters
        ----------
        psf : `PSF`
            PSF to compute an MTF from

        Returns
        -------
        `MTF`
            A new MTF instance

        """
        if getattr(psf, '_mtf', None) is not None:
            return psf._mtf
        else:
            dat = abs(m.fftshift(m.fft2(psf.data)))
            unit_x = forward_ft_unit(psf.sample_spacing / 1e3, psf.samples_x)  # 1e3 for microns => mm
            unit_y = forward_ft_unit(psf.sample_spacing / 1e3, psf.samples_y)
            return MTF(dat / dat[psf.center_y, psf.center_x], unit_x, unit_y)

    @staticmethod
    def from_pupil(pupil, efl, Q=2):
        """Generate an MTF from a pupil, given a focal length (propagation distance).

        Parameters
        ----------
        pupil : `Pupil`
            A pupil to propagate to a PSF, and convert to an MTF
        efl : `float`
            Effective focal length or propagation distance of the wavefunction
        Q : `number`
            ratio of pupil sample count to PSF sample count.  Q > 2 satisfies nyquist

        Returns
        -------
        `MTF`
            A new MTF instance

        """
        psf = PSF.from_pupil(pupil, efl=efl, Q=Q)
        return MTF.from_psf(psf)


def diffraction_limited_mtf(fno, wavelength, frequencies=None, samples=128):
    """Give the diffraction limited MTF for a circular pupil and the given parameters.

    Parameters
    ----------
    fno : `float`
        f/# of the lens.
    wavelength : `float`
        wavelength of light, in microns.
    frequencies : `numpy.ndarray`
        spatial frequencies of interest, in cy/mm if frequencies are given, samples is ignored.
    samples : `int`
        number of points in the output array, if frequencies not given.

    Returns
    -------
    if frequencies not given:
        frequencies : `numpy.ndarray`
            array of ordinate data
        mtf : `numpy.ndarray`
            array of coordinate data
    else:
        mtf : `numpy.ndarray`
            array of MTF data

    Notes
    -----
    If frequencies are given, just returns the MTF.  If frequencies are not
    given, returns both the frequencies and the MTF.

    """
    extinction = 1 / (wavelength / 1000 * fno)
    if frequencies is None:
        normalized_frequency = m.linspace(0, 1, samples)
    else:
        normalized_frequency = m.asarray(frequencies) / extinction
        try:
            normalized_frequency[normalized_frequency > 1] = 1  # clamp values
        except TypeError:  # single freq
            if normalized_frequency > 1:
                normalized_frequency = 1

    mtf = _difflim_mtf_core(normalized_frequency)

    if frequencies is None:
        return normalized_frequency * extinction, mtf
    else:
        return mtf


def _difflim_mtf_core(normalized_frequency):
    """Compute the MTF at a given normalized spatial frequency.

    Parameters
    ----------
    normalized_frequency : `numpy.ndarray`
        normalized frequency; function is defined over [0, and takes a value of 0 for [1,

    Returns
    -------
    `numpy.ndarray`
        The diffraction MTF function at a given normalized spatial frequency

    """
    return (2 / m.pi) * \
           (m.arccos(normalized_frequency) - normalized_frequency *
            m.sqrt(1 - normalized_frequency ** 2))



"""prysm, a python optics module."""
from setuptools import setup

setup(
    name='prysm',
    version='0.14.0',
    description='A python optics module',
    long_description='Uses geometrical and fourier optics to model optical systems',
    license='MIT',
    author='Brandon Dube',
    author_email='brandondube@gmail.com',
    url='https://github.com/brandondube/prysm',
    packages=['prysm'],
    install_requires=['numpy', 'scipy'],
    extras_require={
        'cpu+': ['numba'],
        'cuda': ['cupy'],
        'Mx': ['h5py'],
        'img': ['imageio'],
        'mtf+': ['pandas'],
        'deconv': ['scikit-image'],
        'plotting': ['matplotlib'],
        'deluxe': ['numba', 'cupy', 'h5py', 'iamgeio', 'pandas', 'scikit-image', 'matplotlib'],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
    ]
)



"""Shack Hartmann sensor modeling tools."""
from collections import deque

from .detector import bindown
from .units import waves_to_microns
from .util import share_fig_ax


class ShackHartmann(object):
    """Shack Hartmann Wavefront Sensor.

    Attributes
    ----------
    buffer_depth : `int`
        number of frames to store from captures
    captures : `deque`
        fixed size set of captures
    captures_simple : `deque`
        simple representation of captures; dot grid matricies of spot centroids
    captures_wvl : `deque`
        wavelength associated with each capture
    lenslet_array_shape : `str`, {'square', 'rectangular'}
        shape of lenslets; square will use less than the entire long side of the sensor
    lenslet_efl : `float`
        efl of lenslets, microns
    lenslet_fno : `float`
        F/# of the lenslets
    lenslet_pitch : `float`
        center-to-center spacing of the lenslets
    megapixels : `float`
        total number of megapixels
    num_lenslets : `tuple`
        (x,y) resolution in lenslets
    pixel_locations_x : `numpy.ndarray`
        x locations of pixel centers
    pixel_locations_y : `numpy.ndarray`
        y locations of pixel centers
    pixel_pitch : `float`
        center to center spacing of pixels
    resolution : `tuple`
        (x,y) resolution, px
    sensor_size : `tuple`
        (x,y) sensor size, mm
    total_lenslets : `int`
        total number of lenslets (analagous to megapixels)
    wavelength : `float`
        operating wavelength, microns.  Can change with exposure

    """
    def __init__(self, sensor_size=(36, 24), pixel_pitch=3.99999,
                 lenslet_pitch=375, lenslet_efl=2000, lenslet_fillfactor=0.9,
                 lenslet_array_shape='square', framebuffer=24,
                 wavelength=0.5):
        """Create a new SHWFS object.

        Parameters
        ----------
        sensor_size : `iterable`
            (x, y) sensor sizes in mm
        pixel_pitch : `float`
            center-to-center pixel spacing in um
        lenslet_pitch : `float`
            center-to-center spacing of lenslets in um
        lenslet_efl : `float`
            lenslet focal length, in microns
        lenslet_fillfactor : `float`
            portion of the sensor height filled by the lenslet array.
            0.9 reserves 5% of the height on both the top and bottom of the array
        lenslet_array_shape : `str`, {'square', 'rectangular'}
            square will inscribe a square array within the detector area,
            rectangular will fill the detector area up to the fill factor
        framebuffer : `int`
            maximum number of frames of data to store
        wavelength : `float`
            wavelength of light, microns

        """
        # process lenslet array shape and lenslet offset
        if lenslet_array_shape.lower() == 'square':
            self.lenslet_array_shape = 'square'
        elif lenslet_array_shape.lower() in ('rectangular', 'rect'):
            self.lenslet_array_shape = 'rectangular'
        lenslet_shift = lenslet_pitch // 2

        # store data related to the silicon
        self.sensor_size = sensor_size
        self.pixel_pitch = pixel_pitch
        self.resolution = (int(sensor_size[0] // (pixel_pitch / 1e3)),
                           int(sensor_size[1] // (pixel_pitch / 1e3)))
        self.megapixels = self.resolution[0] * self.resolution[1] / 1e6
        self.pixel_locations_x = None
        self.pixel_locations_y = None

        # compute lenslet array shifts
        if self.lenslet_array_shape == 'square':
            xidx = 1
            sensor_extra_x = sensor_size[0] - sensor_size[1]
            if sensor_extra_x < 0:
                sensor_extra_y = sensor_size[1] - sensor_size[0]
                yshift = (sensor_extra_y / sensor_size[1]) / 2
                xshift = 0
            else:
                xshift = (sensor_extra_x / sensor_size[0]) / 2
                yshift = 0
        else:
            xidx = 0
            xshift = 0
            yshift = 0
        yidx = 1

        # store lenslet metadata - TODO: figure out why I need the - 1
        self.num_lenslets = (int(sensor_size[xidx] * lenslet_fillfactor // (lenslet_pitch / 1e3)) - 1,
                             int(sensor_size[yidx] * lenslet_fillfactor // (lenslet_pitch / 1e3)) - 1)
        self.total_lenslets = self.num_lenslets[0] * self.num_lenslets[1]
        self.lenslet_pitch = lenslet_pitch
        self.lenslet_efl = lenslet_efl
        self.lenslet_fno = self.lenslet_efl / self.lenslet_pitch

        # compute lenslet locations
        start_factor = (1 - lenslet_fillfactor) / 2
        end_factor = 1 - start_factor
        start_factor_x, end_factor_x = start_factor + xshift, end_factor - xshift
        start_factor_y, end_factor_y = start_factor + yshift, end_factor - yshift

        # factors of 1e3 convert mm to um, and round to 0.1nm to avoid machine precision errors
        lenslet_start_x = round(start_factor_x * sensor_size[0] * 1e3 + lenslet_shift, 4)
        lenslet_start_y = round(start_factor_y * sensor_size[1] * 1e3 + lenslet_shift, 4)
        lenslet_end_x = round(end_factor_x * sensor_size[0] * 1e3 - lenslet_shift, 4)
        lenslet_end_y = round(end_factor_y * sensor_size[1] * 1e3 - lenslet_shift, 4)

        lenslet_pos_x = m.linspace(lenslet_start_x, lenslet_end_x, self.num_lenslets[0])
        lenslet_pos_y = m.linspace(lenslet_start_y, lenslet_end_y, self.num_lenslets[0])
        self.refx, self.refy = m.meshgrid(lenslet_pos_x, lenslet_pos_y)

        # initiate the frame buffer and store the wavelength
        self.buffer_depth = framebuffer
        self.captures = deque(maxlen=framebuffer)
        self.captures_simple = deque(maxlen=framebuffer)
        self.captures_wvl = deque(maxlen=framebuffer)
        self.wavelength = wavelength

    def _prep_pixel_grid(self):
        '''Prepare the pixel grid.

        This function allows a new SH WFS object to lazily compute its pixel
        coordinates only when necessary, as a high resolution sensor will take
        a bit of time to prepare.

        Returns
        -------
        self.pixel_locations_x : `numpy.ndarray`
            array of pixel x coordinates, microns
        self.pixel_locations_y : `numpy.ndarray`
            aray of pixel y coordinates, microns

        '''
        if self.pixel_locations_x is None:
            pp = self.pixel_pitch
            pxx, pxy = self.resolution
            pixextx, pixexty = pp * pxx, pp * pxy
            x = m.arange(0, pixextx, pp)
            y = m.arange(0, pixexty, pp)
            self.pixel_locations_x, self.pixel_locations_y = m.meshgrid(x, y)

        return self.pixel_locations_x, self.pixel_locations_y

    def __repr__(self):
        """Represent object.

        Returns
        -------
        `str`
            a string description of this sensor

        """
        return ('Shack Hartmann sensor with: \n'
                f'\t({self.resolution[0]:}x{self.resolution[1]})px, {self.megapixels:.1f}MP CMOS\n'
                f'\t({self.num_lenslets[0]}x{self.num_lenslets[1]})lenslets, '
                f'{self.total_lenslets:1.0f} wavefront samples\n'
                f'\t{self.buffer_depth} frame buffer, currently storing {len(self.captures)} frames')

    def plot_reference_spots(self, fig=None, ax=None):
        """Create a plot of the reference positions of lenslets.

        Parameters
        ----------
        fig : `matplotlib.figure.Figure`, optional
            Figure containing the plot
        ax : `matplotlib.axes.Axis`, optional:
            Axis containing the plot

        Returns
        -------
        fig : `matplotlib.figure.Figure`, optional
            Figure containing the plot
        ax : `matplotlib.axes.Axis`, optional:
            Axis containing the plot

        """
        fig, ax = share_fig_ax(fig, ax)
        ax.scatter(self.refx / 1e3, self.refy / 1e3, c='k', s=8)
        ax.set(xlim=(0, self.sensor_size[0]), xlabel='Detector Position X [mm]',
               ylim=(0, self.sensor_size[1]), ylabel='Detector Position Y [mm]',
               aspect='equal')
        return fig, ax

    def sample_wavefront(self, pupil, make_image=False, fig=None, ax=None):
        """Sample a wavefront, producing a Shack-Hartmann spot grid.

        Parameters
        ----------
        pupil : `Pupil`
            a pupil object
        make_image : `bool`
            boolean, whether to simulate the actual detector; must be false or will raise
            NotImplementedError
        fig : `matplotlib.figure.Figure`, optional
            Figure containing the plot
        ax : `matplotlib.axes.Axis`, optional:
            Axis containing the plot

        Returns
        -------
        psf_centers_x : `numpy.ndarray`
            x coordinate centers of the PSFs
        psf_centers_y : `numpy.ndarray`
            y coordiante centers of the PSFs

        Notes
        -----
        Algorithm is as follows:
            1.  Compute the gradient of the wavefront.
            2.  Bindown the wavefront such that each sample in the output
                corresponds to the local wavefront gradient at a lenslet.
            3.  Compute the x and y delta of each PSF in the image plane.
            4.  Shift each spot by the corresponding delta.
                This is the end for make_image=False.
            5.  If make_image=True, then the dots are painted onto the pixel
                grid and that array is convolved with a sinc function and
                deconvolved with a rect function.  The sinc is the PSF of the
                lenslets, which are assumed to all be the same and independent
                of the local wavefront they sampled, and the rect is to remove
                the finite width of the delta imposed by the pixel grid.

        """
        # grab the phase from the pupil and convert to units of length
        pupil._correct_phase_units()  # -> waves -> microns
        data = pupil.phase * waves_to_microns(pupil.wavelength)
        if pupil.wavelength != self.wavelength:
            self.wavelength = pupil.wavelength

        # convert the phase error to radians in the paraxial approximation
        data /= self.lenslet_pitch  # epd

        # compute the gradient - TODO: see why gradient is dy,dx not dx,dy
        normalized_sample_spacing = 2 / pupil.samples
        dy, dx = m.gradient(data, normalized_sample_spacing, normalized_sample_spacing)

        # convert the gradient from waves to radians -- angle alpha is made as:
        '''
          dz
        ______
        \    |
         \   |
          \ a| rho
           \ |
            \|
             \
            a = tan(dz/rho)
            can either compute "one sided" a, using the vertex to radius as rho
            and the vertex to radius phase error, or compute "two sided" a using
            the epd and the full wavefront error (PV) as dz.
        '''

        # bin to the lenslet area
        nlenslets_y = self.num_lenslets[1]
        npupilsamples_y = pupil.samples
        npx_bin = npupilsamples_y // nlenslets_y
        dx_binned = bindown(dx, npx_bin, mode='avg')
        dy_binned = bindown(dy, npx_bin, mode='avg')

        # compute the lenslet PSF shift
        shift_x, shift_y = psf_shift(self.lenslet_efl, dx_binned, dy_binned)
        psf_centers_x, psf_centers_y = self.refx + shift_x, self.refy + shift_y
        self.captures_simple.append({
            'x': psf_centers_x,
            'y': psf_centers_y})
        self.captures_wvl.append(pupil.wavelength)
        if make_image:
            pass  # TODO: write image simulation
        else:
            self.captures.append(None)
            return psf_centers_x, psf_centers_y

    def plot_simple_result(self, result_index=-1, type='quiver', fig=None, ax=None):
        '''Plot the simple version of the most recent result.

        Parameters
        ----------
        result_index : `int`
            which capture to plot, defaults to most recent
        type : `str`, {'quiver', 'dots'}
            what type of plot to make
        fig : `matplotlib.figure.Figure`, optional
            Figure containing the plot
        ax : `matplotlib.axes.Axis`, optional:
            Axis containing the plot

        Returns
        -------
        fig : `matplotlib.figure.Figure`, optional
            Figure containing the plot
        ax : `matplotlib.axes.Axis`, optional:
            Axis containing the plot

        '''
        idx = result_index

        fig, ax = share_fig_ax(fig, ax)

        mx, my = self.captures_simple[idx]['x'], self.captures_simple[idx]['y']
        if type.lower() in ('q', 'quiver'):
            dx, dy = mx - self.refx, my - self.refy
            ax.quiver(self.refx, self.refy, dx, dy, scale=1, units='xy', scale_units='xy')
        elif type.lower() in ('d', 'dots'):
            ax.scatter(self.refx, self.refy, c='k', s=8)
            ax.scatter(mx, my, c='r', s=8)

        ax.set(xlim=(0, self.sensor_size[0] * 1e3), xlabel=r'Detector X [$\mu m$]',
               ylim=(0, self.sensor_size[1] * 1e3), ylabel=r'Detector Y [$\mu m$]',
               aspect='equal')
        return fig, ax


def psf_shift(lenslet_efl, dx, dy, mag=1):
    '''Compute the shift of a PSF, in microns.

    Parameters
    ----------
    lenslet_efl : `float`
        EFL of lenslets, microns
    dx : `m.ndarray`
        dx gradient of wavefront
    dy : `m.ndarray`
        dy gradient of wavefront
    mag : `float`
        magnification of the collimation system

    Returns
    -------
    `numpy.ndarray`
        array of PSF shifts

    Notes
    -----
    see eq. 12 of Chanan, "Principles of Wavefront Sensing and Reconstruction"
    delta = m * fl * grad(z)
    m is magnification of SH system
    fl is lenslet focal length
    grad(z) is the x, y gradient of the opd, z, which is expressed in radians.

    '''
    coef = -mag * lenslet_efl
    return coef * dx, coef * dy
